""" test sequence

   isort:skip_file
"""
import os
import sys
import unittest
 
SRC = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "src")
if SRC not in sys.path:
    sys.path.insert(0, SRC)

from  cioseq.sequence import Progression, Sequence, _resolve_frames
 


class ResolveFramesTest(unittest.TestCase):
    def test_single_integer(self):
        s = _resolve_frames(1)
        self.assertEqual(s[0], 1)
        self.assertEqual(len(s), 1)

    def test_integer_range_args(self):
        s = _resolve_frames(1, 5)
        self.assertEqual(s[0], 1)
        self.assertEqual(s[4], 5)
        self.assertEqual(len(s), 5)

    def test_integer_range_step_args(self):
        s = _resolve_frames(1, 5, 2)
        self.assertEqual(s[0], 1)
        self.assertEqual(s[2], 5)
        self.assertEqual(len(s), 3)

    def test_sort_integer_range_step_args(self):
        s = _resolve_frames(5, 1, 2)
        self.assertEqual(s[0], 1)
        self.assertEqual(s[2], 5)
        self.assertEqual(len(s), 3)

    def negative_integer_args(self):
        s = _resolve_frames(-5, -1, 2)
        self.assertEqual(s[0], 5)
        self.assertEqual(s[2], 1)
        self.assertEqual(len(s), 3)

    def test_array(self):
        s = _resolve_frames([1, 2, 6, 4])
        self.assertEqual(s[0], 1)
        self.assertEqual(s[2], 4)
        self.assertEqual(len(s), 4)

    def test_single_number_spec(self):
        s = _resolve_frames("3")
        self.assertEqual(s[0], 3)
        self.assertEqual(len(s), 1)

    def test_range_spec(self):
        s = _resolve_frames("1-3")
        self.assertEqual(s[0], 1)
        self.assertEqual(len(s), 3)

    def test_range_step_spec(self):
        s = _resolve_frames("1-5x2")
        self.assertEqual(s[0], 1)
        self.assertEqual(s[2], 5)
        self.assertEqual(len(s), 3)


class SequenceFactoryTest(unittest.TestCase):
    def test_single_number_is_progression(self):
        s = Sequence.create(1)
        self.assertIsInstance(s, Progression)

    def test_range_is_progression(self):
        s = Sequence.create(1, 10)
        self.assertIsInstance(s, Progression)

    def test_range_step_is_progression(self):
        s = Sequence.create(1, 10, 2)
        self.assertIsInstance(s, Progression)

    def test_single_number_string_is_progression(self):
        s = Sequence.create("1")
        self.assertIsInstance(s, Progression)

    def test_range_string_is_progression(self):
        s = Sequence.create("1-10x2")
        self.assertIsInstance(s, Progression)

    def test_frame_spec_is_not_progression(self):
        s = Sequence.create("1-10, 14, 20-50x4")
        self.assertIsInstance(s, Sequence)
        self.assertNotIsInstance(s, Progression)

    def test_list_sequence_is_not_progression(self):
        s = Sequence.create([1, 3, 4, 6, 8, 9])
        self.assertIsInstance(s, Sequence)
        self.assertNotIsInstance(s, Progression)

    def test_progressive_list_is_progression(self):
        s = Sequence.create([1, 3, 5, 7, 9])
        self.assertIsInstance(s, Progression)

    def test_is_progression_method(self):
        s = Sequence.create([1, 3, 5, 7, 9])
        self.assertTrue(s.is_progression())
        s = Sequence.create([1, 3, 5, 7, 9, 10])
        self.assertFalse(s.is_progression())


class SequenceFactoryFailTest(unittest.TestCase):
    def test_negative_step(self):
        with self.assertRaises(ValueError):
            Sequence.create(1, 10, -1)

    def test_bad_spec(self):
        with self.assertRaises(ValueError):
            Sequence.create("f")

    def test_bad_spec_step(self):
        with self.assertRaises(ValueError):
            Sequence.create("1-10xf")


    def test_try_to_bypass_factory(self):
        with self.assertRaises(TypeError):
            Sequence("1-10xf")

    def test_try_to_bypass_factory_with_correct_number_of_args(self):
        with self.assertRaises(AssertionError):
            Sequence(object(), "1-10xf")


class StartEndStepIntsTest(unittest.TestCase):
    def test_create_from_start_only(self):
        s = Sequence.create(1)
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 1)
        self.assertEqual(s.step, 1)

    def test_create_from_start_end_ints(self):
        s = Sequence.create(1, 5)
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 1)

    def test_create_with_step_ints(self):
        s = Sequence.create(1, 5, 2)
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 2)

    def test_create_with_negative_start(self):
        s = Sequence.create(-1, 5, 2)
        self.assertEqual(s.start, -1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 2)

    def test_create_with_negative_start_end(self):
        s = Sequence.create(-5, -1, 2)
        self.assertEqual(s.start, -5)
        self.assertEqual(s.end, -1)
        self.assertEqual(s.step, 2)

    def test_create_swap_negative_start_end(self):
        s = Sequence.create(-1, -5, 2)
        self.assertEqual(s.start, -5)
        self.assertEqual(s.end, -1)
        self.assertEqual(s.step, 2)


class StartEndStepSpecTest(unittest.TestCase):
    def test_create_from_start_end_spec(self):
        s = Sequence.create("1-5")
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 1)

    def test_create_from_start_end_spec_backwards(self):
        s = Sequence.create("5-1")
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 1)

    def test_create_with_step_spec(self):
        s = Sequence.create("1-5x2")
        self.assertEqual(s.start, 1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 2)

    def test_create_with_start_spec(self):
        s = Sequence.create("5")
        self.assertEqual(s.start, 5)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 1)

    def test_create_with_negative_start(self):
        s = Sequence.create("-1-5x2")
        self.assertEqual(s.start, -1)
        self.assertEqual(s.end, 5)
        self.assertEqual(s.step, 2)

    def test_create_with_negative_start_end(self):
        s = Sequence.create("-5--1x2")
        self.assertEqual(s.start, -5)
        self.assertEqual(s.end, -1)
        self.assertEqual(s.step, 2)

    def test_create_swap_negative_start_end(self):
        s = Sequence.create("-1--5x2")
        self.assertEqual(s.start, -5)
        self.assertEqual(s.end, -1)
        self.assertEqual(s.step, 2)


class SequenceToStringTest(unittest.TestCase):
 

    def test_progression_range_step_round_down(self):
        s = Sequence.create(0, 10, 3)
        self.assertEqual(str(s), "0-9x3")

    def test_sequence(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        self.assertEqual(str(s), "1-10,14,20-48x4")

    def test_repr_from_progression(self):
        s = Sequence.create(0, 10, 3)
        self.assertEqual(repr(s), "Sequence.create('0-9x3')")

    def test_repr_from_sequence(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        self.assertEqual(repr(s), "Sequence.create('1-10,14,20-48x4')")


class ChunksTest(unittest.TestCase):
    def test_no_chunk_size(self):
        s = Sequence.create("1-100")
        self.assertEqual(s.chunk_size, 100)

    def test_clamp_chunk_size(self):
        s = Sequence.create("1-100", chunk_size=200)
        self.assertEqual(s.chunk_size, 100)

    def test_chunk_size(self):
        s = Sequence.create("1-100", chunk_size=50)
        self.assertEqual(s.chunk_size, 50)

    def test_chunk_size_setter(self):
        s = Sequence.create("1-100")
        s.chunk_size = 50
        self.assertEqual(s.chunk_size, 50)

    def test_best_chunk_size(self):
        s = Sequence.create("1-100")
        s.chunk_size = 76
        self.assertEqual(s.best_chunk_size(), 50)
        s.chunk_size = 37
        self.assertEqual(s.best_chunk_size(), 34)
        s.chunk_size = 100
        self.assertEqual(s.best_chunk_size(), 100)

    def test_create_chunks_linear(self):
        s = Sequence.create("1-100")
        s.chunk_size = 10
        chunks = s.chunks()
        self.assertEqual(list(chunks[0]), list(range(1, 11)))

    def test_create_chunks_cycle(self):
        s = Sequence.create("1-100")
        s.chunk_size = 10
        s.chunk_strategy = "cycle"
        chunks = s.chunks()
        self.assertEqual(list(chunks[0]), list(range(1, 100, 10)))
        s.chunk_size = 7
        chunks = s.chunks()
        self.assertEqual(list(chunks[0]), list(range(1, 100, 15)))

    def test_cycle_chunks_same_as_cycle_progressions_if_input_progression(self):
        s = Sequence.create("1-100")
        s.chunk_size = 10
        s.chunk_strategy = "cycle"
        c_chunks = s.chunks()
        s.chunk_strategy = "cycle_progressions"
        cp_chunks = s.chunks()
        self.assertEqual( list(c_chunks[0]), list(cp_chunks[0]))
 
    def test_cycle_progression_chunks_are_progressions(self):
        s = Sequence.create("1-1001x3,135-149x2,379,454")
        s.chunk_size = 10
        s.chunk_strategy = "cycle_progressions"
        chunks = s.chunks()
        for chunk in s.chunks():
            self.assertIsInstance(chunk, Progression)
 
    def test_chunk_count(self):
        s = Sequence.create("1-100")
        s.chunk_size = 7
        self.assertEqual(s.chunk_count(), 15)
        s.chunk_size = 15
        self.assertEqual(s.chunk_count(), 7)
        s.chunk_size = 10
        self.assertEqual(s.chunk_count(), 10)

    def test_progression_count_with_chunk_size(self):
        s = Sequence.create("1-3,6-12x2,13,14-36x3")
        s.chunk_size = 4
        s.chunk_strategy = "progressions"
        self.assertEqual(s.chunk_count(), 5)

    def test_progression_count_without_chunk_size(self):
        s = Sequence.create("1-3,6-12x2,13,14-36x3")
        s.chunk_size = -1
        s.chunk_strategy = "progressions"
        self.assertEqual(s.chunk_count(), 4)

    def test_progression_count_all_progressions(self):
        s = Sequence.create("1-3,6-12x2,13,14-36x3")
        s.chunk_size = -1
        s.chunk_strategy = "progressions"
        for chunk in s.chunks():
            self.assertIsInstance(chunk, Progression)


class IntersectionTest(unittest.TestCase):
    def test_does_intersect(self):
        s = Sequence.create("1-10")
        i = s.intersection(range(5, 15))
        self.assertEqual(list(i), list(range(5, 11)))

    def test_does_not_intersect(self):
        s = Sequence.create("1-10")
        i = s.intersection(range(25, 35))
        self.assertEqual(i, None)


class UnionTest(unittest.TestCase):
    def test_creates_union_from_range(self):
        s = Sequence.create("1-10")
        u = s.union(range(5, 15))
        self.assertEqual(list(u), list(range(1, 15)))

    def test_creates_union_from_other_sequence(self):
        s1 = Sequence.create("1-10")
        s2 = Sequence.create("5-15")
        u = s1.union(s2)
        self.assertEqual(list(u), list(Sequence.create("1-15")))

class ChunkIntersectionTest(unittest.TestCase):
    def test_intersecting_chunks(self):
        s = Sequence.create("1-50", chunk_size=5)
        rhs = Sequence.create("1,2,7")
        result = s.intersecting_chunks(rhs)
        self.assertEqual(len(result), 2)
        self.assertEqual( list(result[0]), [1,2,3,4,5])

    def test_intersecting_chunks_when_excess(self):
        s = Sequence.create("1-50", chunk_size=5)
        rhs = Sequence.create("1,2,7,52")
        result = s.intersecting_chunks(rhs)
        self.assertEqual(len(result), 2)
        self.assertEqual( list(result[0]), [1,2,3,4,5])

    def test_intersecting_chunks_when_none(self):
        s = Sequence.create("1-50", chunk_size=5)
        rhs = Sequence.create("52-60")
        result = s.intersecting_chunks(rhs)
        self.assertEqual(len(result), 0)
 
    def test_intersecting_chunks_when_negative(self):
        s = Sequence.create("-1--50", chunk_size=5)
        rhs = Sequence.create("-1,-2,-7,-52")
        result = s.intersecting_chunks(rhs)
        self.assertEqual(len(result), 2)
        self.assertEqual( list(result[0]), [-10, -9, -8, -7,-6])

class SequenceIteratorTest(unittest.TestCase):
    def test_iterator_sorted_no_dups(self):
        s = Sequence.create("1-10, 8-20x2, 19, 17")
        expected = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 14, 16, 17, 18, 19, 20]
        self.assertEqual(list(s), expected)


class ProgressionsTest(unittest.TestCase):
    def test_empty(self):
        result = Progression.factory([])
        self.assertEqual(result, [])

    def test_single(self):
        result = Progression.factory([3])
        self.assertEqual(len(result), 1)
        self.assertEqual(str(result[0]), "3")

    def test_two(self):
        result = Progression.factory([3, 7])
        self.assertEqual(len(result), 1)
        self.assertEqual(str(result[0]), "3-7x4")

    def test_progression_at_start(self):
        result = Progression.factory([3, 5, 7, 10, 15])
        self.assertEqual(len(result), 2)
        self.assertEqual(str(result[0]), "3-7x2")

    def test_progression_at_end(self):
        result = Progression.factory([3, 5, 8, 10, 12])
        self.assertEqual(str(result[1]), "8-12x2")

    def test_order(self):
        numbers = [3, 5, 8, 10, 12]
        numbers.reverse()
        result = Progression.factory(numbers)
        self.assertEqual(str(result[0]), "3-5x2")
        self.assertEqual(str(result[1]), "8-12x2")

    def test_from_range(self):
        result = Progression.factory(xrange(2, 97, 3))
        self.assertEqual(len(result), 1)
        self.assertEqual(len(result[0]), 32)

    def test_range_max_size_one(self):
        result = Progression.factory([2, 4, 6, 8, 10], max_size=1)
        self.assertEqual(len(result), 5)

    def test_range_max_2(self):
        result = Progression.factory([2, 4, 6, 8, 10], max_size=2)
        self.assertEqual(len(result), 3)

    def test_range_max_16(self):
        result = Progression.factory(xrange(2, 197, 3), max_size=16)
        self.assertEqual(len(result[0]), 16)
        self.assertEqual(len(result), 5)

    def test_seq_to_longest_progressions(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        progs = Progression.factory(s)
        self.assertEqual(len(progs), 3)

    def test_seq_to_limited_length_progressions(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        progs = Progression.factory(s, max_size=4)
        self.assertEqual(len(progs), 6)


class PermutationsTest(unittest.TestCase):
    def test_one_substitution(self):

        template = "image.%(frame)04d.tif"
        result = list(Sequence.permutations(template, frame="0-20x2"))
        self.assertIn("image.0008.tif", result)
        self.assertIn("image.0012.tif", result)
        self.assertEqual(len(result), 11)

    def test_two_the_same_substitution(self):

        template = "/path/%(frame)d/image.%(frame)04d.tif"
        result = list(Sequence.permutations(template, frame="0-20x2"))
        self.assertIn("/path/8/image.0008.tif", result)
        self.assertIn("/path/12/image.0012.tif", result)
        self.assertEqual(len(result), 11)

    def test_three_substitutions(self):
        template = "image_%(uval)02d_%(vval)02d.%(frame)04d.tif"
        kw = {"uval": "1-2", "vval": "1-2", "frame": "10-11"}
        result = list(Sequence.permutations(template, **kw))
        self.assertIn("image_01_01.0010.tif", result)
        self.assertIn("image_02_02.0011.tif", result)
        self.assertIn("image_02_01.0010.tif", result)
        self.assertEqual(len(result), 8)


class OffsetTest(unittest.TestCase):
    def test_offset_positive_value(self):
        s = Sequence.create("1-10")
        s = s.offset(5)
        self.assertEqual(list(s), list(range(6, 16)))

    def test_offset_negative_value(self):
        s = Sequence.create("11-20")
        s = s.offset(-5)
        self.assertEqual(list(s), list(range(6, 16)))

    def test_negative_result_offset(self):
        s = Sequence.create("1-10")
        s = s.offset(-3)
        self.assertEqual(list(s), list(range(-2, 8)))

    def test_offset_sequence_has_same_chunk_size(self):
        s = Sequence.create("1-10", chunk_size=3)
        o = s.offset(5)
        self.assertEqual(o.chunk_size, s.chunk_size)

    def test_offset_sequence_has_same_chunk_strategy(self):
        s = Sequence.create("1-10", chunk_strategy="cycle")
        o = s.offset(5)
        self.assertEqual(o.chunk_strategy, s.chunk_strategy)


class ExpandFilenameTest(unittest.TestCase):
    def test_expand_result_same_length(self):
        s = Sequence.create("8-12")
        template = "image.#.exr"
        result = s.expand(template)
        self.assertEqual(len(result), 5)

    def test_expand_single_hash(self):
        s = Sequence.create("8-12")
        template = "image.#.exr"
        result = s.expand(template)
        self.assertIn("image.8.exr", result)
        self.assertIn("image.12.exr", result)

    def test_expand_padded_hash(self):
        s = Sequence.create("8-12")
        template = "image.#####.exr"
        result = s.expand(template)
        self.assertIn("image.00008.exr", result)
        self.assertIn("image.00012.exr", result)

    def test_expand_many_captures(self):
        s = Sequence.create("8-12")
        template = "/some/directory_###/image.#####.exr"
        result = s.expand(template)
        self.assertIn("/some/directory_012/image.00012.exr", result)
        self.assertIn("/some/directory_008/image.00008.exr", result)

    def test_invalid_template_raises(self):
        s = Sequence.create("8-12")
        template = "image.26.exr"
        with self.assertRaises(ValueError):
            s.expand(template)


class ResolveDollarTimeVarsTest(unittest.TestCase):
    def test_resolve_one_file_no_padding(self):
        s = Sequence.create("1")
        template = "image.$F.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.1.exr")

    def test_resolve_no_match_leaves_untouched(self):
        s = Sequence.create("1")
        template = "image.ho.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.ho.exr")

    def test_resolve_many_matches(self):
        s = Sequence.create("1")
        template = "image.$F.$5F.$2F.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.1.00001.01.exr")

    def test_resolve_one_template_to_many_filenames(self):
        s = Sequence.create("1-5")
        template = "image.$2F.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 5)
        self.assertIn("image.02.exr", result)
        self.assertIn("image.05.exr", result)

    def test_resolve_many_templates_to_many_filenames(self):
        s = Sequence.create("1-3")
        templates = [
            "/folder_1/image.$2F.exr",
            "/folder_2/image.$2F.exr",
            "/folder_3/image.$2F.exr",
        ]
        result = s.expand_dollar_f(*templates)
        self.assertEqual(len(result), 3)
        self.assertIn("/folder_2/image.02.exr", result)
        self.assertIn("/folder_3/image.03.exr", result)


class ResolveFormatTest(unittest.TestCase):
    def test_resolve_one_file_no_padding(self):
        s = Sequence.create("1")
        template = "image.{frame}.exr"
        result = s.expand_format(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.1.exr")

    def test_resolve_no_match_leaves_untouched(self):
        s = Sequence.create("1")
        template = "image.ho.exr"
        result = s.expand_format(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.ho.exr")

    def test_resolve_many_matches(self):
        s = Sequence.create("1")
        template = "image.{frame}.{frame:05d}.{frame:02d}.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], "image.1.00001.01.exr")

    def test_resolve_one_template_to_many_filenames(self):
        s = Sequence.create("1-5")
        template = "image.{frame:02d}.exr"
        result = s.expand_dollar_f(template)
        self.assertEqual(len(result), 5)
        self.assertIn("image.02.exr", result)
        self.assertIn("image.05.exr", result)

    def test_resolve_many_templates_to_many_filenames(self):
        s = Sequence.create("1-3")
        templates = [
            "/folder_1/image.{frame:02d}.exr",
            "/folder_2/image.{frame:02d}.exr",
            "/folder_3/image.{frame:02d}.exr",
        ]
        result = s.expand_dollar_f(*templates)
        self.assertEqual(len(result), 3)
        self.assertIn("/folder_2/image.02.exr", result)
        self.assertIn("/folder_3/image.03.exr", result)


class ToCustomSpecTest(unittest.TestCase):
    def test_spec_single_number(self):
        s = Sequence.create(10)
        self.assertEqual(s.to(":", "%", ";"), "10")

    def test_spec_range(self):
        s = Sequence.create(0, 10)
        self.assertEqual(s.to(":", "%", ";"), "0:10")

    def test_spec_step_range(self):
        s = Sequence.create(1, 9, 2)
        self.assertEqual(s.to(":", "%", ";"), "1:9%2")

    def test_spec_complex(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        self.assertEqual(s.to(":", "%", ";"), "1:10;14;20:48%4")

    def test_complex_add_spaces(self):
        s = Sequence.create("1-10, 14, 20-48x4")
        self.assertEqual(s.to(":", "%", "; "), "1:10; 14; 20:48%4")


class SubsampleTest(unittest.TestCase):
    def test_counts_from_1_to_10(self):
        s = Sequence.create("1-10")
        ss = s.subsample(1)
        self.assertEqual(len(ss), 1)
        self.assertEqual(list(ss), [6])
        ss = s.subsample(2)
        self.assertEqual(len(ss), 2)
        self.assertEqual(list(ss), [3, 8])
        ss = s.subsample(3)
        self.assertEqual(len(ss), 3)
        self.assertEqual(list(ss), [2, 6, 9])
        ss = s.subsample(4)
        self.assertEqual(len(ss), 4)
        self.assertEqual(list(ss), [2, 4, 7, 9])
        ss = s.subsample(5)
        self.assertEqual(len(ss), 5)
        self.assertEqual(list(ss), [2, 4, 6, 8, 10])
        ss = s.subsample(6)
        self.assertEqual(len(ss), 6)
        self.assertEqual(list(ss), [1, 3, 5, 6, 8, 10])
        ss = s.subsample(7)
        self.assertEqual(len(ss), 7)
        self.assertEqual(list(ss), [1, 3, 4, 6, 7, 8, 10])
        ss = s.subsample(8)
        self.assertEqual(len(ss), 8)
        self.assertEqual(list(ss), [1, 2, 4, 5, 6, 7, 9, 10])
        ss = s.subsample(9)
        self.assertEqual(len(ss), 9)
        self.assertEqual(list(ss), [1, 2, 3, 4, 6, 7, 8, 9, 10])
        ss = s.subsample(10)
        self.assertEqual(len(ss), 10)
        self.assertEqual(list(ss), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        ss = s.subsample(11)
        self.assertEqual(len(ss), 10)
        self.assertEqual(list(ss), [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


if __name__ == "__main__":
    unittest.main()



"""A Sequence is a set of integers representing frames.

To create a sequence, use the create method and pass either:

frame spec: e.g. string  "1-10, 14, 20-50x4"
range: one, two, or three integers representing start, end, step.
iterable of ints: something that implements __iter__, e.g. a list.

Ranges in a spec are inclusive, e.g. len(Sequence("1-3")) == 3

If the input is valid, either a Sequence or a Progression will
be created. A Progression is a Sequence that can be expressed
as an arithmetic progression, i.e. first,last,step.

A Sequence has methods to split itself into chunks. Chunks are
subsequences and are themselves of type, Sequence or
Progression.

"""
import math
import re
import itertools

RX_FRAME = re.compile(r"\$(\d?)F")

PROGRESSION_SPEC_REGEX = re.compile(
    r"^(?P<first>-?\d+)(-(?P<last>-?\d+)(x(?P<step>[1-9][0-9]*))?)?$"
)

SPLIT_SPEC_REGEX = re.compile(r"[ ,,]+")


def _clamp(minval, val, maxval):
    return sorted([minval, val, maxval])[1]


def _resolve_frames(*args):
    if not args:
        raise TypeError("Need at least one arg")
    frames = []
    if len(args) == 1:
        arg = args[0]
        if hasattr(arg, "__iter__"):
            frames = arg
        else:
            arg = str(arg)
            for progression in SPLIT_SPEC_REGEX.split(arg):
                match = PROGRESSION_SPEC_REGEX.match(progression)
                if not match:
                    raise ValueError("Arg must be 'start<-end<xstep>>")
                first, last, step = [
                    int(match.group("first")),
                    int(match.group("last") or match.group("first")),
                    int(match.group("step") or 1),
                ]
                if step < 1:
                    raise ValueError("Spec must have positive step values")
                first, last = sorted([first, last])
                frames += range(first, last + 1, step)
    else:  # args are inclusive range
        first, last = sorted([int(n) for n in [args[0], args[1]]])
        step = int(args[2]) if len(args) == 3 else 1
        if step < 1:
            raise ValueError("Step arg must be positive")
        frames = xrange(first, last + 1, step)
    return sorted(set(frames))


def _start_end_step(arr):
    """
    Return start, end, step if the array is a progression.
    """
    if len(arr) == 1:
        return (arr[0], arr[0], 1)
    elif len(arr) == 2:
        return (arr[0], arr[1], arr[1] - arr[0])
    elif range(arr[0], arr[-1] + 1, arr[2] - arr[1]) == arr:
        return (arr[0], arr[-1], arr[1] - arr[0])


class Sequence(object):
    """A collection of frames with the ability to generate chunks."""

    __magic_shield = object()

    @staticmethod
    def permutations(template, **kw):
        for vals in itertools.product(
            *(iter(Sequence.create(spec)) for spec in kw.values())
        ):
            subs = dict(zip(kw, vals))
            yield template % subs

    @classmethod
    def create(cls, *args, **kw):
        """Factory which will create either a Sequence or a Progression.

        A Sequence is an arbitrary list of frames with unique sorted
        elements. A Progression is a subclass of Sequence that can be
        expressed as an arithmetic progression: i.e. start, end, step.
        """

        frames = _resolve_frames(*args)
        if not frames:
            raise TypeError("Can't create Sequence with no frames")

        progression_spec = _start_end_step(frames)
        if progression_spec:
            return Progression(cls.__magic_shield, *progression_spec, **kw)
        return Sequence(cls.__magic_shield, frames, **kw)

    def __init__(self, _shield, iterable, **kw):
        """Instantiate from list of ints.

        This method will usually be called by a factory. chunk size
        defaults to the length of the sequence if missing or if -1 is
        given. It is also clamped.
        """
        assert (
            _shield == Sequence.__magic_shield
        ), "Sequence and Progression objects must be created with Sequence.create()"

        self._iterable = iterable

        num = len(self._iterable)
        chunk_size = kw.get("chunk_size", num)
        self._chunk_size = num if chunk_size < 1 else sorted([num, chunk_size])[0]

        self._chunk_strategy = kw.get("chunk_strategy", "linear")

    @property
    def start(self):
        """return the first frame."""
        return self._iterable[0]

    @property
    def end(self):
        """return the last frame."""
        return self._iterable[-1]

    def _cycle_chunks(self):
        """Generate chunks with frame cycling.

        Say we have 100 frames to render (1-100). With chunksize of 5,
        there will be 20 instances. The first chunk will render 1, 21,
        41, 61 81, the second chunk 2, 22, 42, 62, 82 and so on, which
        means the artist will see frames 1,2,3,4...20 early on. This may
        be useful to check for frame coherence artifacts without waiting
        for a whole chunk to render on one machine.
        """
        num_chunks = self.chunk_count()
        result = [[] for i in range(num_chunks)]
        for index, frame in enumerate(self._iterable):
            idx = index % num_chunks
            result[idx].append(frame)
        return [Sequence.create(x) for x in result]

    def _cycle_progression_chunks(self):
        """
        We want cycles which are progressions.

        We know that if we make cycle chunks from a progression, the child
        chunks will themselves be progressions. So all we need to do is split
        the sequence into as few progressions as possible, loop over each
        progression, set the chunk size to our own chunk size, then emit cycle
        chunks.
        """
        chunks = Progression.factory(self._iterable, max_size=-1)

        result = []
        for chunk in chunks:
            chunk.chunk_size = self.chunk_size
            chunk.chunk_strategy = "cycle"
            result += chunk.chunks()
        return result


    def _linear_chunks(self):
        """Generate chunks in sorted order."""
        result = []
        for i in xrange(0, len(self._iterable), self._chunk_size):
            result.append(
                Sequence.create(list(self._iterable)[i : i + self._chunk_size])
            )
        return result

    def chunks(self):
        """return list of chunks according to size and chunk strategy.

        Strategy can be linear, cycle, progressions, or cycle_progressions. 
        Others may be added in future, such as binary.

        "linear" will generate Sequences by simply walking
        along the list of frames and grouping so that each
        group has chunk_size frames. i.e. Fill chunk 1 then chunk 2 ...

        "cycle" will generate Sequences by distributing the
        frames in each chunk in such a way that they get
        filled in parallel group 1 gets frame 1, group 2 gets
        frame 2 and we cycle around. See _cycle_chunks()

        "progressions" tries to walk along as in linear strategy, but
        with the constraint that each Sequence is a Progression and
        can be expressed with start, end, step. See
        Progression.factory()

        "cycle_progressions" makes cycles as described above but
        with the constraint that each Sequence is a Progression.

        """
        if self._chunk_strategy == "cycle":
            return self._cycle_chunks()
        if self._chunk_strategy == "cycle_progressions":
            return self._cycle_progression_chunks()
        if self._chunk_strategy == "cycle_progressions2":
            return self._cycle_progression_chunks2()
        if self._chunk_strategy == "progressions":
            return Progression.factory(self._iterable, max_size=self._chunk_size)
        return self._linear_chunks()

    def chunk_count(self):
        """Calculate the number of chunks that will be emitted.

        If strategy is progressions we just generate them and count the
        objects. Otherwise we can calculate from the frame length and
        chunk size directly
        """
        if self._chunk_strategy == "progressions":
            return len(Progression.factory(self._iterable, max_size=self._chunk_size))
        return int(math.ceil(len(self._iterable) / float(self._chunk_size)))

    def intersection(self, iterable):
        """Generate a Sequence that is the intersection of an iterable with
        this Sequence.

        This is useful for determining which scout frames are valid
        """
        common_frames = set(self._iterable).intersection(set(iterable))
        if not common_frames:
            return None
        return Sequence.create(
            common_frames,
            chunk_size=self._chunk_size,
            chunk_strategy=self._chunk_strategy,
        )


    def intersects(self, iterable):
        """Determine if an iterable intersects this Sequence"""
        return len(set(self._iterable).intersection(set(iterable))) > 0  
 
    def union(self, iterable):
        """Generate a Sequence that is the union of an iterable with this
        Sequence.

        Useful for getting a sequence that covers multiple
        output ranges.
        """
        union_frames = set(self._iterable).union(set(iterable))

        return Sequence.create(
            union_frames,
            chunk_size=self._chunk_size,
            chunk_strategy=self._chunk_strategy,
        )

    def offset(self, value):
        """Generate a new Sequence with all values offset.

        It is an error if the new sequence would contain negative
        numbers.
        """
        offset_frames = [x + value for x in self._iterable]
        return Sequence.create(
            offset_frames,
            chunk_size=self._chunk_size,
            chunk_strategy=self._chunk_strategy,
        )

    def intersecting_chunks(self, iterable):
        """Return chunks that intersect the given sequence."""
        result = []
        for c in self.chunks():
            if c.intersects(iterable):
                result.append(c)
        return result


    def expand(self, template):
        """Expand a hash template with this sequence.

        Example /some/directory_###/image.#####.exr. Sequence is invalid
        if it contains no hashes. First we replace the hashes with a
        template placeholder that specifies the padding as a number.
        Then we use format to replace the placehoders with numbers.
        """
        if not re.compile("#+").search(template):
            raise ValueError("Template must contain hashes.")

        format_template = re.sub(
            r"(#+)", lambda match: "{{0:0{:d}d}}".format(len(match.group(1))), template
        )

        return [format_template.format(el) for el in self._iterable]

    def expand_format(self, *templates):
        """
        cycle the templates until there are as many as there are frames
        in the sequence. Then replace with the frame
        """
        result = []
        for f, template in zip(self._iterable, itertools.cycle(templates)):
            result.append(template.format(frame=f))
        return result

    def expand_dollar_f(self, *templates):
        """
        Expands $ templates such as those containing $3F or $F.

        If a single template is given, such as image.$2F.exr, and the sequence
        contains [1,2,4] then the result will be:
        [
            "image.01.exr",
            "image.02.exr",
            "image.04.exr"
        ]
        However, if there are 3 templates, such as:
        a_1/image.$2F.exr
        a_2/image.$2F.exr
        a_4/image.$2F.exr
        then the result will be:
        [
            a_1/image.01.exr,
            a_2/image.02.exr,
            a_4/image.04.exr
        ]
        i.e. there will always be len(sequence) elements in the result.

        The intention is that the number of templates given be either 1 or
        len(sequence), but for completeness, we cycle if there is some number
        between, and clip if there are too many.
        """
        result = []
        for f, template in zip(self._iterable, itertools.cycle(templates)):
            format_template = re.sub(
                r"\$(\d?)F",
                lambda match: "{{frame:0{:d}d}}".format(int(match.group(1) or 0)),
                template,
            )
            result.append(format_template.format(frame=f))
        return result

    def subsample(self, count):
        """Take a selection of elements from the sequence.

        Return value is a new sequence where the elements are plucked in
        the most distributed way.
        """
        n = len(self)
        count = _clamp(1, count, n)

        res = []
        gap = n / float(count)
        pos = gap / 2.0

        for _ in range(count):
            index = int(pos)
            res.append(self._iterable[index])
            pos += gap

        spec = ",".join([str(x) for x in res])
        return Sequence.create(spec)

    def best_chunk_size(self):
        """Determine the best distribution of frames per chunk based on the
        current number of chunks.

        For example, if chunk size is 70 and there are 100 frames, 2
        chunks will be generated with 70 and 30 frames. It would be
        better to have 50 frames per chunk and this method returns that
        number. NOTE: It doesn't make sense to use the best chunk size
        when chunk strategy is progressions.
        """
        count = int(math.ceil(len(self._iterable) / float(self._chunk_size)))
        return int(math.ceil(len(self._iterable) / float(count)))

    def is_progression(self):
        """Is this sequence a progression."""
        return isinstance(self, Progression)

    def to(self, range_sep, step_sep, block_sep):
        return (
            str(self)
            .replace("-", range_sep)
            .replace("x", step_sep)
            .replace(",", block_sep)
        )

    @property
    def chunk_size(self):
        """Return chunk size."""
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, value):
        """Set chunk_size.

        Max is the length of frames and min is 1. If a value less than 1
        is given, set it to max, which means the default is to output 1
        chunk.
        """
        num = len(self._iterable)
        self._chunk_size = num if value < 1 else sorted([num, value])[0]

    @property
    def chunk_strategy(self):
        """Return the current strategy for emitting chunks."""
        return self._chunk_strategy

    @chunk_strategy.setter
    def chunk_strategy(self, value):
        """Set strategy for emitting chunks."""
        self._chunk_strategy = value

    def __iter__(self):
        return iter(self._iterable)

    def __len__(self):
        return len(self._iterable)

    def __str__(self):
        """String representation contains the stringified progressions."""
        progs = Progression.factory(self._iterable)
        return (",").join([str(p) for p in progs])

    def __repr__(self):
        """Repr contains whats necessary to recreate."""
        return "Sequence.create(%r)" % (str(self))


class Progression(Sequence):
    def __init__(self, _shield, start, end, step, **kw):
        self._iterable = xrange(start, end + 1, step)
        self.chunk_size = kw.get("chunk_size", -1)
        self._chunk_strategy = kw.get("chunk_strategy", "linear")

    @property
    def step(self):
        if self.start == self.end:
            return 1
        return self._iterable[1] - self._iterable[0]

    @property
    def range(self):
        return (self.start, self.end, self.step)

    def __str__(self):
        if self.start == self.end:
            return str(self.start)
        if self.step == 1:
            return "%s-%s" % (self.start, self.end)
        return "%s-%sx%s" % (self.start, self.end, self.step)

    @staticmethod
    def factory(iterable, **kw):
        """Split a sequence of integers into arithmetic progressions.

        This is not necessarily the most compact set of progressions in
        the sequence as that would take too long for large sets, and be
        really difficult to code. This algorithm walks through the
        sorted sequence, gathers elements with the same gap as the
        previous element. The max size keyword arg will limit the length
        of progressions.
        """
        max_size = kw.get("max_size", len(iterable))
        iterable = sorted(iterable)

        if max_size == 1:
            return [Sequence.create(prog) for prog in iterable]

        if max_size < 1:
            max_size = len(iterable)

        results = [[]]

        # add a sentinel to the end - see below
        iterable.append(-1)
        prog = 0
        for element in iterable:
            # if not adding to current progression, create a new progression.
            if not _should_add(element, results[prog], max_size):
                last_prog = prog
                prog += 1
                results.append([])

                # if the last progression only has 2 elements, then steal
                # the second of them to start this progression. Why? because
                # we are greedy and will have more chance of making longer
                # progressions if we dismantle progressions of length 2 and
                # then pair up the straggling singles later.
                # Note that the sentinel added above ensures this rule also
                # works on the original last element.
                if len(results[last_prog]) == 2:
                    results[prog].append(results[last_prog][1])
                    results[last_prog] = results[last_prog][:1]

            results[prog].append(element)

        # remove the sentinel from the last progression,
        # or remove the whole last
        # progression if it contains only the sentinel.
        if len(results[prog]) == 1:
            results = results[:-1]
        else:
            results[prog] = results[prog][:-1]

        # Now join straggling singles into pairs if possible. Loop
        # through the singles, joining every other one to the one
        # before. In theory we could recurse here - looking for
        # longer progressions in the stragglers.
        singles = []
        j = 0
        for i in range(len(results)):
            if len(results[j]) == 1:
                singles.append(results.pop(j)[0])
            else:
                j += 1

        for i, single in enumerate(singles):
            if i % 2 == 0:
                results.append([single])
            else:
                results[-1].append(single)

        results.sort(key=lambda v: v[0])

        return [Sequence.create(prg) for prg in results]


def _should_add(element, progression, max_size):
    """Should an element be added to a progression?

    Specifically, if the progression has 0 or 1 elements, then add it,
    as this starts the progression. Otherwise check if the gap between
    the element and the last in the progression is the same as the gap
    in the rest of the progression.
    """
    if len(progression) < 2:
        return True
    if len(progression) == max_size:
        return False
    return progression[1] - progression[0] == element - progression[-1]

